/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Peers facility.
 *
 * This file contains the code to send events to peers.
 * The reception of events is done inside the peer state machine (peer-psm.c)
 *
 */

#include "waaad-internal.h"
#include "peer-internal.h"

EVENTSTR_DEFINITION;

/* Send an event to a peer from inside; the peer lock must be held. */
int _peer_events_send_nl(_peer_t * peer, _pevent_t event, void * data)
{
	_pe_t * pe = NULL;
	
	TRACE_ENTRY("%p(%s) %d(%s) %p", peer, (peer && peer->p_diamid)?peer->p_diamid:"???", event, EVENTSTR(event), data);
	
	CHECK_PARAMS(  VALIDATE_PEER( peer ) && (event < PEVENT_MAX ));
	
	/* create the new event */
	CHECK_MALLOC(  pe = (_pe_t *) malloc(sizeof(_pe_t))  );
	memset(pe, 0, sizeof(_pe_t));
	uti_list_init(&pe->list, data);
	pe->event = event;
	
	/* enqueue the new event */
	uti_list_insert_before(&peer->p_events, _LIST(pe));
	
	/* signal the condition */
	CHECK_POSIX(  pthread_cond_signal(&peer->p_condvar)  );
	
	return 0;
}

/* Send an event to a peer from outside; the peer lock must not be held. */
/* Periodical check for cancelation state when trying to lock the mutex. */
int _peer_events_send(_peer_t * peer, _pevent_t event, void * data)
{
	int ret = 0;
	
	CHECK_PARAMS(  VALIDATE_PEER( peer )  );
	
	/* loop until we suceed to lock the peer. */
	do {
		struct timespec	to;
		CHECK_SYS(  clock_gettime(CLOCK_REALTIME, &to)  );
		to.tv_sec += 1; /* one second timeout */
		
		pthread_testcancel();
		
		ret = pthread_mutex_timedlock( &peer->p_lock, &to );
	} while (ret == ETIMEDOUT);
	
	CHECK_POSIX(ret);
	
	pthread_cleanup_push(cleanup_mutex, &peer->p_lock);

	/* Create and queue the event */
	CHECK_FCT_DO(  ret = _peer_events_send_nl(peer, event, data), /* continue */  );
	
	/* unlock */
	pthread_cleanup_pop(0);
	CHECK_POSIX(  pthread_mutex_unlock(&peer->p_lock)  );
	
	return ret;
}

void _peer_events_dispose_data( _pevent_t event, void ** data )
{
	CHECK_PARAMS_DO(data, return);
	
	switch (event) {
		case PEVENT_INCNX:
			{
				_ev_incoming_t * evdata = (_ev_incoming_t *)(*data);
				_peer_struct_destroy(&evdata->peer);
			}
			break;

		case PEVENT_CONNECTED:
			{
				_ev_connected_t * evdata = (_ev_connected_t *)(*data);
				CHECK_SYS_DO(  shutdown(evdata->sock, SHUT_RDWR),  /* continue */);
			}
			break;

		default:
			/* for other events, nothing to do */
			break;
	}
	free(*data);
	*data = NULL;
	return;
}

/* reset the received events */
void _peer_events_reset( _peer_t * peer )
{
	while (!IS_LIST_EMPTY(&peer->p_events)) {
		_pe_t * pe = (_pe_t *) (peer->p_events.next);
		
		uti_list_unlink(_LIST(pe));
		
		TRACE_DEBUG(INFO, "Discarding event %s (%d)", EVENTSTR(pe->event), pe->event);
		
		/* we have to free pe->list.o here also */
		_peer_events_dispose_data(pe->event, &pe->list.o);
		
		/* free the event */
		free(pe);
	}
}	

