/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Messages queues module.
 * 
 * This module provides the support functions for queueing and picking messages.
 * See queues.h for more information.
 *
 * The threads that call these functions must be in the cancellation state PTHREAD_CANCEL_ENABLE and type PTHREAD_CANCEL_DEFERRED.
 * This is the default state and type on thread creation.
 *
 * In order to destroy properly a queue, the application must:
 *  -> shutdown any process that can add into the queue first.
 *  -> pthread_cancel any thread that could be waiting on the queue.
 *  -> consume any message that is in the queue, using meq_tryget.
 *  -> then destroy the queue using meq_del.
 */

#include "waaad-internal.h"

/********************************************************************************************************/

/* Internal representation of a queue */
typedef struct {
	int		eyec;	/* An eye catcher, also used to check a queue is valid. _MEQ_EYEC */
	int		count;	/* number of objects in the list */
	int		thrs;	/* number of threads waiting for objects */
	uti_list_t	list;	/* sentinel for the list */
	pthread_mutex_t	mtx;	/* Mutex protecting this queue */
	pthread_cond_t	cond;	/* condition variable of the list */
} _meq_t;

/* The eye catcher value */
#define _MEQ_EYEC	0xe7ec1130

/* Macro to cast */
#define _Q( _ptr ) ((_meq_t *)(_ptr))

/* Macro to check a pointer */
#define CHECK_QUEUE( _queue ) (( (_queue) != NULL) && ( _Q(_queue)->eyec == _MEQ_EYEC) )


/* The global queues pointers are actually declared here */
meq_t * g_meq_incoming = NULL;	/* Received messages */
meq_t * g_meq_outgoing = NULL;	/* Messages to send */
meq_t * g_meq_local = NULL;	/* Messages for local handling */

/********************************************************************************************************/

/* Initialize the module */
int meq_init ( void )
{
	TRACE_ENTRY( );
	
	/* Initialize the global queues */
	CHECK_FCT(  meq_new(&g_meq_incoming)  );
	CHECK_FCT(  meq_new(&g_meq_outgoing)  );
	CHECK_FCT(  meq_new(&g_meq_local)  );
	
	return 0;
}

/* End of the module */
int meq_fini ( void )
{
	TRACE_ENTRY( );
	
	/* Destroy the global queues */
	CHECK_FCT(  meq_del(&g_meq_incoming)  );
	CHECK_FCT(  meq_del(&g_meq_outgoing)  );
	CHECK_FCT(  meq_del(&g_meq_local)  );

	return 0;
}

/* Create a new queue */
int meq_new ( meq_t ** queue )
{
	int ret = 0;
	_meq_t * new;
	
	TRACE_ENTRY( "%p", queue );
	
	CHECK_PARAMS( queue );
	
	/* Create a new object */
	CHECK_MALLOC( new = (_meq_t *) malloc (sizeof (_meq_t) )  );
	
	/* Initialize the content */
	memset(new, 0, sizeof(_meq_t));
	
	new->eyec = _MEQ_EYEC;
	uti_list_init(&new->list, NULL);
	
	CHECK_POSIX_DO( ret = pthread_mutex_init(&new->mtx, NULL),  { free(new); return ret; }  );
		
	CHECK_POSIX_DO( ret = pthread_cond_init(&new->cond, NULL),  { free(new); return ret; }  );
	
	/* We're done */
	*queue = (meq_t *) new;
	return 0;
}

/* Delete a queue */
int meq_del ( meq_t ** queue )
{
	TRACE_ENTRY( "%p", queue );

	CHECK_PARAMS( CHECK_QUEUE( *queue ) );
	
	CHECK_POSIX(  pthread_mutex_lock( &_Q(*queue)->mtx )  );
	
	if ((_Q(*queue)->count != 0) || (_Q(*queue)->thrs != 0)) {
		TRACE_DEBUG(INFO, "The queue count is %d, thread count %d, cannot destroy", _Q(*queue)->count, _Q(*queue)->thrs);
		CHECK_POSIX_DO(  pthread_mutex_unlock( &_Q(*queue)->mtx ), /* no fallback */  );
		return EINVAL;
	}
	
	/* sanity check */
	ASSERT(IS_LIST_EMPTY(&_Q(*queue)->list));
	
	/* Ok, now invalidate the queue */
	_Q(*queue)->eyec = 0xdead;
	
	/* And destroy it */
	CHECK_POSIX(  pthread_mutex_unlock( &_Q(*queue)->mtx )  );
	
	CHECK_POSIX(  pthread_cond_destroy( &_Q(*queue)->cond )  );
	
	CHECK_POSIX(  pthread_mutex_destroy( &_Q(*queue)->mtx )  );
	
	free(*queue);
	*queue = NULL;
	
	return 0;
}

/* Retrieve the number of messages of a queue */
int meq_length ( meq_t * queue, int * length )
{
	TRACE_ENTRY( "%p %p", queue, length );
	
	/* Check the parameters */
	CHECK_PARAMS( CHECK_QUEUE( queue ) && length );
	
	/* lock the queue */
	CHECK_POSIX(  pthread_mutex_lock( &_Q(queue)->mtx )  );
	
	/* Retrieve the count */
	*length = _Q(queue)->count;
	
	/* Unlock */
	CHECK_POSIX(  pthread_mutex_unlock( &_Q(queue)->mtx )  );
	
	/* Done */
	return 0;
}
int meq_length_noerr ( meq_t * queue )
{
	if ( !CHECK_QUEUE( queue ) )
		return 0;
	return _Q(queue)->count;
}

/* Enqueue a message */
int meq_post ( meq_t * queue, msg_t * msg )
{
	uti_list_t * new;
	
	TRACE_ENTRY( "%p %p", queue, msg );
	
	/* Check the parameters */
	CHECK_PARAMS( CHECK_QUEUE( queue ) && msg );
	
	/* Create a new list item */
	CHECK_MALLOC(  new = (uti_list_t *) malloc (sizeof (uti_list_t))  );
	
	uti_list_init(new, (void *)msg);
	
	/* lock the queue */
	CHECK_POSIX(  pthread_mutex_lock( &_Q(queue)->mtx )  );
	
	/* Add the new message at the end */
	uti_list_insert_before( &_Q(queue)->list, new);
	_Q(queue)->count++;
	
	/* Signal if threads are asleep */
	if (_Q(queue)->thrs > 0) {
		CHECK_POSIX(  pthread_cond_signal(&_Q(queue)->cond)  );
	}
	
	/* Unlock */
	CHECK_POSIX(  pthread_mutex_unlock( &_Q(queue)->mtx )  );
	
	/* Done */
	return 0;
}

/* Pop the first message from the queue */
static msg_t * _meq_pop(meq_t * queue)
{
	void * ret = NULL;
	uti_list_t * li;
	
	ASSERT( ! IS_LIST_EMPTY(&_Q(queue)->list) );
	
	uti_list_unlink(li = _Q(queue)->list.next);
	_Q(queue)->count--;
	ret = li->o;
	free(li);
	
	return (msg_t *)ret;
}

/* Retrieve a message, if any */
int meq_tryget ( meq_t * queue, msg_t ** msg )
{
	int wouldblock = 0;
	
	TRACE_ENTRY( "%p %p", queue, msg );
	
	/* Check the parameters */
	CHECK_PARAMS( CHECK_QUEUE( queue ) && msg );
	
	/* lock the queue */
	CHECK_POSIX(  pthread_mutex_lock( &_Q(queue)->mtx )  );
	
	/* Check queue status */
	if (_Q(queue)->count > 0) {
		/* There are messages in the queue, so pick the first one */
		*msg = _meq_pop(queue);
	} else {
		wouldblock = 1;
		*msg = NULL;
	}
		
	/* Unlock */
	CHECK_POSIX(  pthread_mutex_unlock( &_Q(queue)->mtx )  );
	
	/* Done */
	return wouldblock ? EWOULDBLOCK : 0;
}

/* This handler is called when a thread is blocked on a queue, and cancelled */
static void _meq_cleanup(void * queue)
{
	TRACE_ENTRY( "%p", queue );
	
	/* Check the parameter */
	if ( ! CHECK_QUEUE( queue )) {
		TRACE_DEBUG(INFO, "Invalid queue, skipping handler");
		return;
	}
	
	/* The thread has been cancelled, therefore it does not wait on the queue anymore */
	_Q(queue)->thrs--;
	
	/* Now unlock the queue, and we're done */
	CHECK_POSIX_DO(  pthread_mutex_unlock( &_Q(queue)->mtx ),  /* nothing */  );
	
	/* End of cleanup handler */
	return;
}

/* The internal function for meq_timedget and meq_get */
static int _meq_tget ( _meq_t * queue, msg_t ** msg, int istimed, const struct timespec *abstime)
{
	int ret = 0;
	int timedout = 0;
	
	/* Check the parameters */
	CHECK_PARAMS( CHECK_QUEUE( queue ) && msg && (abstime || !istimed) );
	
	/* Initialize the msg value */
	*msg = NULL;
	
	/* lock the queue */
	CHECK_POSIX(  pthread_mutex_lock( &queue->mtx )  );
	
awaken:
	/* Check queue status */
	if (queue->count > 0) {
		/* There are messages in the queue, so pick the first one */
		*msg = _meq_pop(queue);
	} else {
		/* We have to wait for a new message */
		queue->thrs++ ;
		pthread_cleanup_push(_meq_cleanup, queue);
		if (istimed) {
			ret = pthread_cond_timedwait( &queue->cond, &queue->mtx, abstime );
		} else {
			ret = pthread_cond_wait( &queue->cond, &queue->mtx );
		}
		pthread_cleanup_pop(0);
		queue->thrs-- ;
		if (ret == 0)
			goto awaken;  /* test for spurious wake-ups */
		
		if (istimed && (ret == ETIMEDOUT)) {
			timedout = 1;
		} else {
			/* Unexpected error condition (means we need to debug) */
			CHECK_POSIX( ret );
		}
	}
	
	/* Unlock */
	CHECK_POSIX(  pthread_mutex_unlock( &queue->mtx )  );
	
	/* Done */
	return timedout ? ETIMEDOUT : 0;
}

	
/* Retrieve a message */
int meq_get ( meq_t * queue, msg_t ** msg )
{
	TRACE_ENTRY( "%p %p", queue, msg );
	return _meq_tget(_Q(queue), msg, 0, NULL);
}

/* Retrieve a message, if any. wait only a certain time */
int meq_timedget ( meq_t * queue, msg_t ** msg, const struct timespec *abstime )
{
	TRACE_ENTRY( "%p %p %p", queue, msg, abstime );
	return _meq_tget(_Q(queue), msg, 1, abstime);
}




